-- Author      : Skylar State
-- Last Updated : 9/21/2014 04:42 AM
local UnitIsConnected = UnitIsConnected
local tinsert, tremove, twipe = table.insert, table.remove, table.wipe
local friendlySpells, resSpells, longEnemySpells, enemySpells, petSpells = {}, {}, {}, {}, {}


local SkyUI = SkyUI
local SkyBaseClass = SkyBaseClass
local cfg = SkyUI.Config.UnitFrames

local debug

local SkyUnitFrame = SkyUI_CreateClass(SkyBaseClass)

local rangeframe

SkyUnitFrame.prototype.F = nil -- Frame

SkyUnitFrame.prototype.RT = 0 -- RelativeTo Position
SkyUnitFrame.prototype.X = 0
SkyUnitFrame.prototype.Y = 0
SkyUnitFrame.prototype.W = 0
SkyUnitFrame.prototype.H = 0

SkyUnitFrame.prototype.Unit = nil
SkyUnitFrame.prototype.GUID = nil
SkyUnitFrame.prototype.Health = 0
SkyUnitFrame.prototype.Power = 0

SkyUnitFrame.prototype.ExtraUnits = {}

SkyUnitFrame.prototype.TargetGUID = nil
SkyUnitFrame.prototype.TexturesReversed = nil

SkyUnitFrame.prototype.Spec = nil
SkyUnitFrame.prototype.TS = nil
SkyUnitFrame.prototype.tO = {X = 0, Y = 0}
SkyUnitFrame.prototype.Txt = nil

SkyUnitFrame.prototype.FeaturesActive = {Focus = false, TargetTarget = false, Pet = true, TargetPet = true, PetTarget = false}

SkyUnitFrame.prototype.IsCasting = false
SkyUnitFrame.prototype.CastIsChanneled = false
SkyUnitFrame.prototype.CastEnd = nil
SkyUnitFrame.prototype.CastDuration = nil
SkyUnitFrame.prototype.Spell = nil

SkyUnitFrame.prototype.IncomingHeals = 0

SkyUnitFrame.prototype.AuraBars = nil
SkyUnitFrame.prototype.AuraBarCount = 0

local UpdateInterval = 0.033
SkyUnitFrame.prototype.TimeSinceLastUpdate = 0

local function updateTooltip(self)
 if( GameTooltip:IsOwned(self) ) then
  if self.TT.Index == nil then GameTooltip:SetUnit(self.TT.Unit)
  else GameTooltip:SetUnitAura(self.TT.Unit, self.TT.Index, self.TT.Filter) end
 end
end

local function showTooltip(self)
 GameTooltip:SetOwner(self, "ANCHOR_BOTTOMLEFT")
 if self.TT.Index == nil then GameTooltip:SetUnit(self.TT.Unit)
 else GameTooltip:SetUnitAura(self.TT.Unit, self.TT.Index, self.TT.Filter) end

 self:SetScript("OnUpdate", updateTooltip)
end

local function hideTooltip(self)
 self:SetScript("OnUpdate", nil)
 GameTooltip:Hide()
end

local function DisableClickable(f, u)
 if not f.F[u .. "Clickable"] or InCombatLockdown() then return end
 f.F[u .. "Clickable"]:EnableMouse(false);
 f.F[u .. "Clickable"].TT = { Unit = nil };
 f.F[u .. "Clickable"]:SetScript("OnEnter", nil)
 f.F[u .. "Clickable"]:SetScript("OnLeave", nil)
 f.F[u .. "Clickable"]:Hide();
end
local function EnableClickable(f, u)
 if not f.F[u .. "Clickable"] or InCombatLockdown() then return end
 f.F[u .. "Clickable"]:EnableMouse(true);
 f.F[u .. "Clickable"].TT = { Unit = u };
 f.F[u .. "Clickable"]:SetScript("OnEnter", showTooltip)
 f.F[u .. "Clickable"]:SetScript("OnLeave", hideTooltip)
 f.F[u .. "Clickable"]:Show();
end

local function AddClickableRegion(f, u, r) -- frame, unit, region
 f.F[u .. "Clickable"] = CreateFrame("Button", SkyUI.BaseName .. f.Unit .. u .. "Clickable", UIParent, "SecureUnitButtonTemplate")
 f.F[u .. "Clickable"]:EnableMouse(true)
 f.F[u .. "Clickable"]:RegisterForClicks("AnyUp")
 f.F[u .. "Clickable"]:SetAttribute("type", "target")
 f.F[u .. "Clickable"]:SetAttribute("unit", u)
 f.F[u .. "Clickable"]:SetSize(r:GetSize())
 f.F[u .. "Clickable"]:SetPoint(r:GetPoint())
 
-- if cfg[u].Enabled then EnableClickable(f, u)
-- else DisableClickable(f, u) end

end


-- f.F["pet" .. "Clickable"]:EnableMouse(false);
-- f.F["pettarget" .. "Clickable"]:EnableMouse(false);
-- f.F["focus" .. "Clickable"]:EnableMouse(false);
-- f.F["targettarget" .. "Clickable"]:EnableMouse(false);


local function Round(num, idp) return tonumber(string.format("%." .. (idp or 0) .. "f", num)) end
local function AddCommas(amt) local str = amt while true do str, k = string.gsub(str, "^(-?%d+)(%d%d%d)", '%1,%2') if (k==0) then break end end return str end

function SkyUnitFrame.prototype:init(unit)
 SkyUI[unit] = self;

 local frameSpec = SkyUI.Specs.UnitFrames.Frames[unit:gsub("^%l", string.upper)]
--	if unit == "player" then frameSpec = SkyUI.Specs.UnitFrames.Frames.Player
--	elseif unit == "target" then frameSpec = SkyUI.Specs.UnitFrames.Frames.Target end
 self.Name = unit
	self.RT = frameSpec.RT
	self.X = frameSpec.X
	self.Y = frameSpec.Y
	self.W = frameSpec.W
	self.H = frameSpec.H
	self.T = SkyUI.Specs.UnitFrames.Unit.Textures
	self.Txt = SkyUI.Specs.UnitFrames.Unit.Text
	self.AuraBars = {}

 self.ExtraUnits = {
  Pet = { Health = 1, Power = 1 },
  PetTarget = { Health = 1, Power = 1 },
  TargetOrFocus = { Health = 1, Power = 1 }
 }

	self.TS = SkyUI.Specs.UnitFrames.TextureSheet
	if SkyUI.Specs.UnitFrames.TextureOffset then self.tO = SkyUI.Specs.UnitFrames.TextureOffset end	
	
 self.Unit = unit

	self.TargetGUID = UnitGUID(self.Unit .. "target")

 self:CreateFrame()
	
	if unit == "target" then self.TexturesReversed = true end
 self:AddTextures()
	
 self.F.Clickable = CreateFrame("Button", SkyUI.BaseName .. self.Unit .. "Clickable", self.F, "SecureUnitButtonTemplate")
 self.F.Clickable:EnableMouse(true)
 self.F.Clickable:RegisterForClicks("AnyUp")
 self.F.Clickable:SetAttribute("type1", "target")
 self.F.Clickable:SetAttribute("type2", "menu")
 self.F.Clickable:SetAttribute("unit", self.Unit)
 self.F.Clickable:SetAllPoints(self.F)
 if self.Unit == "player" then self.F.Clickable.menu = function(self) ToggleDropDownMenu(1, 1, PlayerFrameDropDown, self, 0 ,0) end
 elseif self.Unit == "target" then self.F.Clickable.menu = function(self) ToggleDropDownMenu(1, 1, TargetFrameDropDown, self, 0 ,0) end end

 self.F.Clickable:SetScript("OnEnter", showTooltip)
 self.F.Clickable:SetScript("OnLeave", hideTooltip)
 self.F.Clickable.TT = { Unit = self.Unit }


 AddClickableRegion(self, self.Unit .. "pet", self.F.PetBG)
 AddClickableRegion(self, self.Unit .. "pettarget", self.F.PetTargetBG)
 
 if self.Unit == "player" then AddClickableRegion(self, "focus", self.F.TargetOrFocusBG)
 elseif self.Unit == "target" then AddClickableRegion(self, "targettarget", self.F.TargetOrFocusBG) end

	self.F.CastBar:Hide()
	self.F.Cast:Hide()
	self.F.CastIconOverlay:Hide()
	
 if self.Txt then self:AddText() end -- Add any frame text/labels if provided

-- self:SetText_Justified("PetName", UnitName(self.Unit .. "pet"), UnitLevel(self.Unit .. "pet"), "LEFT", 75)
-- self:SetText_Justified("UnitHealth", self:AddCommas(UnitHealth(unit)) .. "/" .. self:AddCommas(UnitHealthMax(unit)) .. " " .. round(UnitHealth(unit) / UnitHealthMax(unit) * 100, 0) .. '%', nil, "TOPLEFT", nil) -- frame, text, level, justification, maxlength
-- self:SetText_Justified("UnitPower", self:AddCommas(UnitPower(unit)) .. "/" .. self:AddCommas(UnitPowerMax(unit)), nil, "TOPLEFT", nil) -- frame, text, level, justification, maxlength
--	self.F.UnitName:SetText(UnitName(self.Unit))
--	self:UpdateUnitName(self.Unit)
--	self.F.UnitName:SetAlpha(0)
--	self.F.CDSpannerL:SetAlpha(1)
--	self.F.CDSpannerS:SetAlpha(1)
--	self.F.CDSpannerR:SetAlpha(1)
-- self.F.UnitLevel:SetText(UnitLevel(unit))

 self.F.Portrait = self:CreatePortrait(self.Unit, "Portrait", 39, 5, -4)

 self.F.ClassIcon = self.F:CreateTexture(SkyUI.BaseName .. self.Unit .. "ClassIcon", "OVERLAY")
	self.F.ClassIcon:SetSize(24, 24)
	self.F.ClassIcon:SetTexture("Interface\\TargetingFrame\\UI-Classes-Circles")
 if self.TexturesReversed then
	 self.F.ClassIcon:SetPoint("TOPRIGHT", self.F, -43, -22)
	else self.F.ClassIcon:SetPoint("TOPLEFT", self.F, 43, -22) end


 self.F.PetPortrait = self:CreatePortrait(self.Unit .. "pet", "PetPortrait", 29, self.T.PetBG.X + self.tO.X + 4, self.T.PetBG.Y + self.tO.Y - 3)
 self.F.PetTargetPortrait = self:CreatePortrait(self.Unit .. "pettarget", "PetTargetPortrait", 29, self.T.PetTargetBG.X + self.tO.X + 4, self.T.PetTargetBG.Y + self.tO.Y - 3)
 if self.Unit == "player" then self.F.TargetOrFocusPortrait = self:CreatePortrait(self.Unit .. "focus", "FocusPortrait", 29, self.T.TargetOrFocusBG.X + self.tO.X + 4, self.T.TargetOrFocusBG.Y + self.tO.Y - 3)
 else self.F.TargetOrFocusPortrait = self:CreatePortrait(self.Unit .. "target", "TargetTargetPortrait", 29, self.T.TargetOrFocusBG.X + self.tO.X + 4, self.T.TargetOrFocusBG.Y + self.tO.Y - 3) end
 

 self:CreateAuraTextures()
 self:AuraUpdates()
 self:UpdateVertBar(self.F.IncomingHealth, 1, UnitHealthMax(self.Unit), self.T.IncomingHealth)


 self.F:SetScript("OnUpdate", self.Update)
 self:OnConfigUpdated()
 if unit == "target" and not UnitExists("target") then self:Hide() end
end


--function SkyUnitFrame.prototype:OnEnter(motion) end
--function SkyUnitFrame.prototype:OnLeave(motion) end

--/run local m,f=GetMouseFocus(),CnB or CreateFrame("Button","CnB",nil,"SecureActionButtonTemplate") f:SetAttribute("type","cancelaura") f:SetAttribute("unit","player") f:SetAttribute("spell",m and m._buff and m._buff.name)
--/click CnB





local function AddSpell(table, spellID)
  local name = GetSpellInfo(spellID)
  if name then
    local usable, nomana = IsUsableSpell(name)
    if usable or nomana then
      table[#table + 1] = name
    end
  end
end

local _,class = UnitClass("player")
local function UpdateSpellList()
  twipe(friendlySpells)
  twipe(resSpells)
  twipe(longEnemySpells)
  twipe(enemySpells)
  twipe(petSpells)
  
  if class == "PRIEST" then
    AddSpell(enemySpells, 585) -- Smite
    AddSpell(longEnemySpells, 589) -- Shadow Word: Pain
    AddSpell(friendlySpells, 2061) -- Flash Heal
    AddSpell(resSpells, 2006) -- Resurrection
  elseif class == "DRUID" then
    AddSpell(enemySpells, 33786) -- Cyclone
    AddSpell(longEnemySpells, 5176) -- Wrath
    AddSpell(friendlySpells, 774) -- Rejuvenation
    AddSpell(resSpells, 50769) -- Revive 
    AddSpell(resSpells, 20484) -- Rebirth 
  elseif class == "PALADIN" then
    AddSpell(enemySpells, 20271) -- Judgement
    AddSpell(friendlySpells, 85673) -- Word of Glory
    AddSpell(resSpells, 7328) -- Redemption
    AddSpell(longEnemySpells, 114165) -- Holy Prism
    AddSpell(longEnemySpells, 114157) -- Execution Sentence
  elseif class == "SHAMAN" then
    AddSpell(enemySpells, 8042) -- Earth Shock 
    AddSpell(longEnemySpells, 403) -- Lightning Bolt
    AddSpell(friendlySpells, 8004) -- Healing Surge
    AddSpell(resSpells, 2008) -- Ancestral Spirit 
  elseif class == "WARLOCK" then
    AddSpell(enemySpells, 5782) -- Fear
    AddSpell(longEnemySpells, 172) -- Corruption
    AddSpell(longEnemySpells, 686) -- Shadow Bolt
    AddSpell(longEnemySpells, 17962) -- Conflag
    AddSpell(petSpells, 755) -- Health Funnel
    AddSpell(friendlySpells, 5697) -- Unending Breath
  elseif class == "MAGE" then
    AddSpell(enemySpells, 118) -- Polymorph
    AddSpell(longEnemySpells, 44614) -- Frostfire Bolt
    AddSpell(friendlySpells, 475) -- Remove Curse
  elseif class == "HUNTER" then
    AddSpell(petSpells, 136) -- Mend Pet
    AddSpell(enemySpells, 75) -- Auto Shot
  elseif class == "DEATHKNIGHT" then
    AddSpell(enemySpells, 49576) -- Death Grip
    AddSpell(friendlySpells, 47541) -- Death Coil
    AddSpell(resSpells, 61999) -- Raise Ally 
  elseif class == "ROGUE" then
    AddSpell(enemySpells, 2094) -- Blind 
    AddSpell(longEnemySpells, 1725) -- Distract
    AddSpell(friendlySpells, 57934) -- Tricks of the Trade
  elseif class == "WARRIOR" then
    AddSpell(enemySpells, 5246) -- Intimidating Shout
    AddSpell(enemySpells, 100) -- Charge
    AddSpell(longEnemySpells, 355) -- Taunt
  elseif class == "MONK" then
    AddSpell(enemySpells, 115546) -- Provoke
    AddSpell(friendlySpells, 115450) -- Detox
    AddSpell(resSpells, 115178) -- Resuscitate
  end 
end

local function getUnit(unit)
  if not unit:find("party") or not unit:find("raid") then
    for i=1, 4 do
      if UnitIsUnit(unit, "party"..i) then
        return "party"..i
      end
    end

    for i=1, 40 do
      if UnitIsUnit(unit, "raid"..i) then
        return "raid"..i
      end
    end
  else
    return unit
  end
end

local function friendlyIsInRange(unit)
  if CheckInteractDistance(unit, 2) and UnitInPhase(unit) then --Inspect (28 yards) and same phase as you
    return true
  end

  if UnitIsDeadOrGhost(unit) and #resSpells > 0 then
    for _, name in ipairs(resSpells) do
      if IsSpellInRange(name, unit) == 1 then
        return true
      end
    end

    return false
  end

  if #friendlySpells == 0 and (UnitInRaid(unit) or UnitInParty(unit)) then
    unit = getUnit(unit)
    return unit and UnitInRange(unit)
  else
    for _, name in ipairs(friendlySpells) do
      if IsSpellInRange(name, unit) == 1 then
        return true
      end
    end
  end
  
  return false
end

local function petIsInRange(unit)
  if CheckInteractDistance(unit, 2) then
    return true
  end
  
  for _, name in ipairs(friendlySpells) do
    if IsSpellInRange(name, unit) == 1 then
      return true
    end
  end
  for _, name in ipairs(petSpells) do
    if IsSpellInRange(name, unit) == 1 then
      return true
    end
  end
  
  return false
end

local function enemyIsInRange(unit)
  if CheckInteractDistance(unit, 2) then
    return true
  end
  
  for _, name in ipairs(enemySpells) do
    if IsSpellInRange(name, unit) == 1 then
      return true
    end
  end
  
  return false
end

local function enemyIsInLongRange(unit)
  for _, name in ipairs(longEnemySpells) do
    if IsSpellInRange(name, unit) == 1 then
      return true
    end
  end
  
  return false
end





function SkyUnitFrame.prototype:CreatePortrait(unit, name, size, posX, posY)

 t = self.F:CreateTexture(SkyUI.BaseName .. unit .. "Portrait", "ARTWORK")
 t:SetSize(size, size)
 if self.TexturesReversed then
  t:SetPoint("TOPRIGHT", self.F, -posX, posY)
  t:SetTexCoord(1, 0, 1, 1, 0, 0, 0, 1)
 else t:SetPoint("TOPLEFT", self.F, posX, posY) end
-- t:SetDrawLayer("OVERLAY", 7)
--  self.F.Portrait:SetTexCoord(0.10, 0.90, 0.10, 0.90)
  local result = SetPortraitTexture(t, unit)
 return t
end

function SkyUnitFrame.prototype:CreateAuraFrame(n, i, t)

 local f = CreateFrame("Frame", SkyUI.BaseName .. self.Unit .. n .. "Frame" .. i, self.F)
--	f:EnableMouse(true)
	
 if self.TexturesReversed then t.RT = "TOPRIGHT"
 else t.RT = "TOPLEFT" end

 if self.TexturesReversed then f:SetPoint(t.RT, -t.X - self.tO.X, t.Y + self.tO.Y)
 else f:SetPoint("TOPLEFT", t.X + self.tO.X, t.Y + self.tO.Y) end
 f:SetSize(t.W, t.H)
 f:SetFrameLevel(5)
		
 local a = f:CreateTexture()
 a:SetAllPoints(f)
--	a:SetTexture(1, 0, 0, 1)
 a:SetTexCoord(0.1, 0.9, 0.1, 0.9)
 f.texture = a
 
 f:SetScript("OnEnter", showTooltip)
 f:SetScript("OnLeave", hideTooltip)
 
 f.TT = nil
 
 local v = { String = "", Font = SkyUI.Fonts.Marken, Size = 8, Flags = "OUTLINE", Clr = SkyUI.Clr.None }

 if t.Txt then
  local txt = f:CreateFontString(SkyUI.BaseName .. self.Name .. n .. "Frame" .. i .. "txt", "OVERLAY")
  txt:SetFont(v.Font, v.Size, v.Flags)
  txt:ClearAllPoints()
  txt:SetText(v.String)

  local bg = f:CreateTexture(SkyUI.BaseName .. self.Name .. n .. "Frame" .. i .. "txt" .. "bg", "BACKGROUND")
  bg:ClearAllPoints()

  bg:SetSize(txt:GetStringWidth() + 10, 14)
  bg:SetTexture(0, 0, 0, 0.7)
  
  
  if self.TexturesReversed then
   txt:SetPoint(t.RT, -13, -2)

   bg:SetPoint(t.RT, f, "TOPRIGHT", -15, 1)
  else
   txt:SetPoint(t.RT, 15, -2)
   txt:SetJustifyH("RIGHT")

   bg:SetPoint(t.RT, f, "TOPLEFT", 15, 1)
  end

  
  self.F[n .. "Frame" .. i .. "txt"] = txt
  self.F[n .. "Frame" .. i .. "txtBG"] = bg

 end
 
 self.F[n .. "Frame" .. i] = f

 f:Hide()
end

function SkyUnitFrame.prototype:CreateAuraTextures()
 local a, ad, b, d, bb, db, t = self.T.AuraBuff, self.T.AuraDebuff, self.T.Buff, self.T.Debuff, self.T.BuffBar, self.T.DebuffBar, {}
	
--	for i = 0, 12, 1 do
--	 local ii = i + 1
--  t["AuraBG" .. ii] = { X = 39 - 20 * (i % 2) - 20 * (math.floor(i / 2)), Y = -90.5 + 20 * (math.floor(i / 2)), tX = 182, tY = 374, W = 21, H = 21, L = "BACKGROUND" }
--	end

 for i=1, 12 do
	local ii = i - 1
  t["BuffBar" .. i] = { X = bb.X, Y = bb.Y + ii * 17, tX = bb.tX, tY = bb.tY, W = bb.W, H = bb.H, L = bb.L, Clr = bb.Clr }
  t["BuffLife" .. i] = { X = b.X, Y = b.Y + ii * 17, tX = b.tX, tY = b.tY, W = b.W, H = b.H, L = b.L, Clr = b.Clr }
	self:CreateAuraFrame("Buff", i, { X = a.X + 1, Y = a.Y + 17 + ii * 17 - 1, tX = a.tX, tY = a.tY, W = a.W - 2, H = a.H - 2, L = "OVERLAY", Clr = SkyUI.Clr.Transparent, Txt = true })
--  t["BuffIcon" .. i] = 
  t["BuffIconBG" .. i] = { X = a.X, Y = a.Y + 17 + ii * 17, tX = a.tX, tY = a.tY, W = a.W, H = a.H, L = "BACKGROUND", Clr = a.Clr }


  t["DebuffBar" .. i] = { X = db.X, Y = db.Y - ii * 17, tX = db.tX, tY = db.tY, W = db.W, H = db.H, L = db.L, Clr = db.Clr }
  t["DebuffLife" .. i] = { X = d.X, Y = d.Y - ii * 17, tX = d.tX, tY = d.tY, W = d.W, H = d.H, L = d.L, Clr = SkyUI.Clr.Debuff, Clr = d.Clr }
	self:CreateAuraFrame("Debuff", i, { X = ad.X + 1, Y = ad.Y - 16 - ii * 17 - 1, tX = ad.tX, tY = ad.tY, W = ad.W - 2, H = ad.H - 2, L = "OVERLAY", Clr = SkyUI.Clr.Transparent, Txt = true })
  t["DebuffIconBG" .. i] = { X = ad.X, Y = ad.Y - 16 - ii * 17, tX = ad.tX, tY = ad.tY, W = ad.W, H = ad.H, L = "BACKGROUND", Clr = ad.Clr }

--  t["DebuffIcon" .. i] = 
-- end

-- for i=1, 12 do
--	local ii = i - 1
  t["AuraBuff" .. i] = { X = a.X + (math.floor(ii / 2) * 17), Y = a.Y - ((ii % 2) * 17), tX = a.tX, tY = a.tY, W = a.W, H = a.H, L = a.L }
	self:CreateAuraFrame("AuraBuff", i, { X = a.X + (math.floor(ii / 2) * 17) + 1, Y = a.Y - ((ii % 2) * 17) - 1, tX = a.tX, tY = a.tY, W = a.W - 2, H = a.H - 2, L = "OVERLAY", Clr = SkyUI.Clr.Transparent })
--  t["AuraBuffIcon" .. i] = 

  t["AuraDebuff" .. i] = { X = ad.X + (math.floor(ii / 2) * 17), Y = ad.Y + ((ii % 2) * 17), tX = ad.tX, tY = ad.tY, W = ad.W, H = ad.H, L = ad.L }
	self:CreateAuraFrame("AuraDebuff", i, { X = ad.X + (math.floor(ii / 2) * 17) + 1, Y = ad.Y + ((ii % 2) * 17) - 1, tX = ad.tX, tY = ad.tY, W = ad.W - 2, H = ad.H - 2, L = "OVERLAY", Clr = SkyUI.Clr.Transparent })
--  t["AuraDebuffIcon" .. i] = 
 end
	
 self:AddTextures(t)

-- for i=1, 8 do self.F["DebuffLife" .. db]:SetVertexColor(.5, 0, 1) -- THIS NEEDS TO BE MOVED TO BUFF FRAME INIT!

-- self:CreateAuraFrames("BuffIcon", 8)
-- self:CreateAuraFrames("DebuffIcon", 8)
-- self:CreateAuraFrames("AuraBuffIcon", 12)
-- self:CreateAuraFrames("AuraDebuffIcon", 12)

end

function SkyUnitFrame.prototype:UpdateVertBar(t, curVal, maxVal, v, YOffset) -- ...
 if curVal == 0 then curVal = 1 end
 if maxVal == 0 or t == nil then return end
	if curVal > maxVal then curVal = maxVal end
	if not YOffset then YOffset = 0 end
	
	local p = curVal / maxVal;

	local newHeight = v.H * p
	local dif = v.H - newHeight

  local g = self.TS
	
	local x1, x2, y1, y2 = v.tX / g.W, (v.tX + v.W) / g.W, (v.tY+dif) / g.H, (v.tY + v.H) / g.H

  point, relativeTo, relativePoint, xOfs, yOfs = t:GetPoint()

	if self.TexturesReversed then
	 t:SetTexCoord(x2, y1, x2, y2, x1, y1, x1, y2)
 	t:SetPoint(relativePoint, t:GetParent(), -v.X - self.tO.X, v.Y + self.tO.Y - dif + YOffset)
	else
	 t:SetTexCoord(x1, x2, y1, y2)
 	t:SetPoint(relativePoint, t:GetParent(), v.X + self.tO.X, v.Y + self.tO.Y - dif + YOffset)
	end
	
	t:SetSize(v.W, newHeight)
end


--  if sui.ExtraUnits.Pet.Health ~= extraUnits.Pet.Health then sui:UpdateHorzBar(self.PetHealth, extraUnits.Pet.Health, UnitHealthMax(sui.Unit .. "pet"), sui.T.PetHealth) end
--  if sui.ExtraUnits.Pet.Power ~= extraUnits.Pet.Power then sui:UpdateHorzBar(self.PetPower, extraUnits.Pet.Power, UnitPowerMax(sui.Unit .. "pet"), sui.T.PetPower) end


function SkyUnitFrame.prototype:UpdateHorzBar(t, curVal, maxVal, v, XOffset) -- , textureOffset, reversedTexture
 if curVal == 0 or not curVal then curVal = 0.1 end
 if maxVal == 0 or t == nil then return end
 if curVal > maxVal then curVal = maxVal end
 if not XOffset then XOffset = 0 end
 
 local p = curVal / maxVal;

 local newWidth = v.W * p
 local dif = v.W - newWidth

 local g = self.TS

-- local x1, x2, y1, y2 = v.tX / g.W,     (v.tX + v.W) / g.W,     (v.tY+dif) / g.H,     (v.tY + v.H) / g.H
 local x1, x2, y1, y2 = v.tX / g.W,     (v.tX + v.W -dif) / g.W,     v.tY / g.H,     (v.tY + v.H) / g.H
 
 point, relativeTo, relativePoint, xOfs, yOfs = t:GetPoint()

 if self.TexturesReversed then
  t:SetTexCoord(x2, y1, x2, y2, x1, y1, x1, y2)
  t:SetPoint(relativePoint, t:GetParent(), -v.X - self.tO.X, v.Y + self.tO.Y)
 else
  t:SetTexCoord(x1, x2, y1, y2)
  t:SetPoint(relativePoint, t:GetParent(), v.X + self.tO.X, v.Y + self.tO.Y)
 end

-- t:SetTexCoord(x1, x2, y1, y2)
-- if self.TexturesReversed then
--  t:SetPoint(relativePoint, -v.X - self.tO.X + XOffset, v.Y + self.tO.Y)
-- else
--  t:SetPoint(relativePoint, v.X + self.tO.X + XOffset, v.Y + self.tO.Y)
-- end
 
 
 
 
 t:SetSize(newWidth, v.H)
end









function SkyUnitFrame.prototype:UpdateUnitReaction()
 
 if cfg.ColorOptions.Health == SkyUI.Structs.By.Class or 1 == 1 then
  local class, classFileName = UnitClass(self.Unit)
  local c = RAID_CLASS_COLORS[classFileName]
  if c then self.F.Health:SetVertexColor(c.r, c.g, c.b) end
 else
  if self.Unit == "target" and UnitCanAttack("player", "target") and (UnitReaction("player", "target") < 4 or (UnitReaction("player", "target")  == 4 and UnitThreatSituation("player", "target"))) then
   self.F.Health:SetVertexColor(SkyUI.Clr.HealthHostile.r, SkyUI.Clr.HealthHostile.g, SkyUI.Clr.HealthHostile.b, SkyUI.Clr.HealthHostile.a)
   self.F.IncomingHealth:SetVertexColor(SkyUI.Clr.HealthHostileIncoming.r, SkyUI.Clr.HealthHostileIncoming.g, SkyUI.Clr.HealthHostileIncoming.b, SkyUI.Clr.HealthHostileIncoming.a)
   self.F.UnitHealth:SetVertexColor(SkyUI.Clr.HealthHostileIncoming.r, SkyUI.Clr.HealthHostileIncoming.g, SkyUI.Clr.HealthHostileIncoming.b, SkyUI.Clr.HealthHostileIncoming.a)
  else
   self.F.Health:SetVertexColor(SkyUI.Clr.HealthFriendly.r, SkyUI.Clr.HealthFriendly.g, SkyUI.Clr.HealthFriendly.b, SkyUI.Clr.HealthFriendly.a)
   self.F.IncomingHealth:SetVertexColor(SkyUI.Clr.HealthIncoming.r, SkyUI.Clr.HealthIncoming.g, SkyUI.Clr.HealthIncoming.b, SkyUI.Clr.HealthIncoming.a)
   if UnitIsPlayer(self.Unit) then self.F.UnitHealth:SetVertexColor(SkyUI.Clr.HealthFriendly.r, SkyUI.Clr.HealthFriendly.g, SkyUI.Clr.HealthFriendly.b, SkyUI.Clr.HealthFriendly.a)
   else self.F.UnitHealth:SetVertexColor(SkyUI.Clr.HealthNeutral.r, SkyUI.Clr.HealthNeutral.g, SkyUI.Clr.HealthNeutral.b, SkyUI.Clr.HealthNeutral.a) end
  end
 end
end

function SkyUnitFrame.prototype:UnitReaction_Other(unit1, unit2, t)
 if UnitCanAttack(unit1, unit2) and (UnitReaction(unit1, unit2) < 4 or (UnitReaction(unit1, unit2)  == 4 and UnitThreatSituation(unit1, unit2))) then
  t:SetVertexColor(SkyUI.Clr.HealthHostile.r, SkyUI.Clr.HealthHostile.g, SkyUI.Clr.HealthHostile.b, SkyUI.Clr.HealthHostile.a)
 else
  t:SetVertexColor(SkyUI.Clr.HealthFriendly.r, SkyUI.Clr.HealthFriendly.g, SkyUI.Clr.HealthFriendly.b, SkyUI.Clr.HealthFriendly.a)
 end 
end

function SkyUnitFrame.prototype:Instantiate(self)
  self:UpdateUnitName()
  self:UpdateUnitReaction()
  self:UpdateClassColor()
  
  self:CastStop(evt, self.Unit)
  if(UnitCastingInfo(self.Unit)) then self:CastStart() end
  
  self:AuraUpdates()
  self:UpdateUnitFrames()
  SetPortraitTexture(self.F.Portrait, self.Unit)
  
  self:UpdateDisplayPower()

  if UnitIsPlayer(self.Unit) then
   local t = CLASS_ICON_TCOORDS[select(2,UnitClassBase(self.Unit))]
   if t then self.F.ClassIcon:SetTexture("Interface\\TargetingFrame\\UI-Classes-Circles") self.F.ClassIcon:SetTexCoord(unpack(t)) end
  else self.F.ClassIcon:SetTexCoord(0,0,0,0) end
  
  self.Health = -1
  self.Power = -1

end

local function UpdateStat(s, stat, statVal, percentVal) -- self, stat, stat amount, stat percent
 local statText, percentText, spanner, leftCap = s.F["Unit" .. stat], s.F["Unit" .. stat .. "Percent"], stat .. "SpannerS", stat .. "SpannerL"
 local percentAlpha, percentWidth = percentText:GetAlpha(), 0
-- if percentText:GetAlpha() == 0 then end
 if percentAlpha > 0 then
  percentText:SetText(percentVal .. '%')
  percentWidth = percentText:GetStringWidth()
 else percentWidth = 0 if statVal == 0 then statVal = "-" end end
-- if not (statVal == "-") then statVal = "(" .. statVal .. ")" end 
 statText:SetText(AddCommas(statVal))
 
 local totalWidth = statText:GetStringWidth() + percentWidth - 4
 
 s.F[spanner]:SetWidth(totalWidth)
 
 local xO, yO = 0, s.T[spanner].Y + s.tO.Y
 if s.TexturesReversed then
  xO = -s.T[spanner].X - s.tO.X + totalWidth - 2
--  left:SetPoint("TOPRIGHT", -leftSpec.X - self.tO.X + hw, leftSpec.Y + self.tO.Y)
  s.F[leftCap]:SetPoint("TOPRIGHT", xO + s.T[leftCap].W, yO)
  s.F[spanner]:SetPoint("TOPRIGHT", xO, yO)
  statText:SetPoint("TOPRIGHT", xO + 3, yO - 4.5)
  if percentAlpha > 0 then percentText:SetPoint("TOPRIGHT", -s.T[spanner].X - s.tO.X + percentWidth + 3 - 5, yO - 4.5) end
 else
  xO = s.T[spanner].X + s.tO.X - totalWidth + 1
--  left:SetPoint("TOPLEFT", leftSpec.X + self.tO.X - hw, leftSpec.Y + self.tO.Y)
  s.F[leftCap]:SetPoint("TOPLEFT", xO - s.T[leftCap].W, yO)
  s.F[spanner]:SetPoint("TOPLEFT", xO, yO)
  statText:SetPoint("TOPLEFT", xO, yO - 4.5)
  if percentAlpha > 0 then percentText:SetPoint("TOPLEFT", s.T[spanner].X + s.tO.X - percentWidth + 5, yO - 4.5) end
 end
 
end


-- **************  UPDATE ! UPDATE ! UPDATE ! UPDATE ! UPDATE ! UPDATE ! UPDATE ! UPDATE ! UPDATE ! UPDATE ! UPDATE ! ********************************** ! UPDATE ! UPDATE --
function SkyUnitFrame.prototype:Update(elapsed)
 local sui = self.sui

 if sui.Spell then sui:CastUpdate() end


 if sui.TimeSinceLastUpdate < UpdateInterval then
  sui.TimeSinceLastUpdate = sui.TimeSinceLastUpdate + elapsed
		return
	end
	
 sui.TimeSinceLastUpdate = 0
	
	local guid = UnitGUID(sui.Unit)
	local targetGUID = UnitGUID(sui.Unit .. "target")
	local health = UnitHealth(sui.Unit)
	local power = UnitPower(sui.Unit)

 if guid then sui:UpdateAuraBars()
 else sui.GUID = guid self:SetAlpha(0) return true end


 local tof = "focus"
 if sui.Unit == "target" then tof = "targettarget" end
 local extraUnits = { Pet = { Health = 1, Power = 1 }, PetTarget = { Health = 1, Power = 1 }, TargetOrFocus = { Health = 1, Power = 1 } }

 if UnitExists(sui.Unit .. "pet") and cfg[sui.Unit .. "pet"].Enabled then extraUnits.Pet = { Health = UnitHealth(sui.Unit .. "pet"), Power = UnitPower(sui.Unit .. "pet") } end
 if UnitExists(sui.Unit .. "pettarget") and cfg[sui.Unit .. "pettarget"].Enabled then extraUnits.PetTarget = { Health = UnitHealth(sui.Unit .. "pettarget"), Power = UnitPower(sui.Unit .. "pettarget") } end
 if UnitExists(tof) and cfg[tof].Enabled then extraUnits.TargetOrFocus = { Health = UnitHealth(tof), Power = UnitPower(tof) } end
 
	-- Either target has changed or we haven't setup the player frame yet
	if( guid ~= sui.GUID ) then
  if not sui.GUID then if sui.Unit == "target" then self:SetAlpha(1) else self:SetAlpha(cfg.UI.player.OOCAlpha) end end -- this frame was previously hidden, let's show it again.
	 sui:Instantiate(sui) -- Update frame for new unit
	end


 -- Check old values against new ones and update
	if sui.Health ~= health then
	 sui:UpdateVertBar(self.Health, health, UnitHealthMax(sui.Unit), sui.T.Health)
	 sui:IncomingHealsUpdate(nil, sui.Unit)
  UpdateStat(sui, "Health", UnitHealth(sui.Unit), Round(UnitHealth(sui.Unit) * 100 / UnitHealthMax(sui.Unit)))
	end
 if sui.Power ~= power then sui:UpdateVertBar(self.Power, power, UnitPowerMax(sui.Unit), sui.T.Power)
  UpdateStat(sui, "Power", UnitPower(sui.Unit), Round(UnitPower(sui.Unit) * 100 / UnitPowerMax(sui.Unit)))
 end



 if UnitExists(sui.Unit .. "pet") and cfg[sui.Unit .. "pet"].Enabled then
  sui:UnitReaction_Other("player", sui.Unit .. "pet", self.PetHealth)
  if sui.ExtraUnits.Pet.Health ~= extraUnits.Pet.Health then sui:UpdateHorzBar(self.PetHealth, extraUnits.Pet.Health, UnitHealthMax(sui.Unit .. "pet"), sui.T.PetHealth) end
  if sui.ExtraUnits.Pet.Power ~= extraUnits.Pet.Power then sui:UpdateHorzBar(self.PetPower, extraUnits.Pet.Power, UnitPowerMax(sui.Unit .. "pet"), sui.T.PetPower) end
  sui.ExtraUnits.Pet.Health = extraUnits.Pet.Health sui.ExtraUnits.Pet.Power = extraUnits.Pet.Power
 end
  
 if UnitExists(sui.Unit .. "pettarget") and cfg[sui.Unit .. "pettarget"].Enabled then
  sui:UnitReaction_Other(sui.Unit .. "pet", sui.Unit .. "pettarget", self.PetTargetHealth)
  if sui.ExtraUnits.PetTarget.Health ~= extraUnits.PetTarget.Health then sui:UpdateHorzBar(self.PetTargetHealth, extraUnits.PetTarget.Health, UnitHealthMax(sui.Unit .. "pettarget"), sui.T.PetTargetHealth) end
  if sui.ExtraUnits.PetTarget.Power ~= extraUnits.PetTarget.Power then sui:UpdateHorzBar(self.PetTargetPower, extraUnits.PetTarget.Power, UnitPowerMax(sui.Unit .. "pettarget"), sui.T.PetTargetPower) end
  sui.ExtraUnits.PetTarget.Health = extraUnits.PetTarget.Health sui.ExtraUnits.PetTarget.Power = extraUnits.PetTarget.Power
 end
 
 if UnitExists(tof) and cfg[tof].Enabled then
  sui:UnitReaction_Other(sui.Unit, tof, self.TargetOrFocusHealth)
  if sui.ExtraUnits.TargetOrFocus.Health ~= extraUnits.TargetOrFocus.Health then sui:UpdateHorzBar(self.TargetOrFocusHealth, extraUnits.TargetOrFocus.Health, UnitHealthMax(tof), sui.T.TargetOrFocusHealth) end
  if sui.ExtraUnits.TargetOrFocus.Power ~= extraUnits.TargetOrFocus.Power then sui:UpdateHorzBar(self.TargetOrFocusPower, extraUnits.TargetOrFocus.Power, UnitPowerMax(tof), sui.T.TargetOrFocusPower) end
  sui.ExtraUnits.TargetOrFocus.Health = extraUnits.TargetOrFocus.Health sui.ExtraUnits.TargetOrFocus.Power = extraUnits.TargetOrFocus.Power
 end  

	sui.GUID = guid
	sui.TargetGUID = targetGUID
	sui.Health = health
	sui.Power = power

 if cfg.RangeTransparency and sui.Unit == "target" then
  local range = { insideAlpha = cfg.UI.target.InsideAlpha, outsideAlpha = cfg.UI.target.OutsideAlpha }
  if UnitCanAttack("player", "target") then
    if enemyIsInRange("target") then self:SetAlpha(range.insideAlpha)
    elseif enemyIsInLongRange("target") then self:SetAlpha(range.insideAlpha)
    else self:SetAlpha(range.outsideAlpha) end
  elseif UnitIsUnit("target", "pet") then
    if petIsInRange("pet") then self:SetAlpha(range.insideAlpha)
    else self:SetAlpha(range.outsideAlpha) end
  else
    if friendlyIsInRange("target") and UnitIsConnected("target") then self:SetAlpha(range.insideAlpha)
    else self:SetAlpha(range.outsideAlpha) end
  end
 end

end

function SkyUnitFrame.prototype:CastStart(evt, ...)
	-- START ... currentTime .. END
	-- (END - currentTime) / Duration = Percent Left
 	
	local spell, rank, displayName, icon, startTime, endTime, isTradeSkill = UnitCastingInfo(self.Unit)
	if not (spell) then
	 spell, rank, displayName, icon, startTime, endTime = UnitChannelInfo(self.Unit)
	 self.CastIsChanneled = true;
	end
 if not (spell) then return end
 self.IsCasting = true

	self.Spell = spell
	self.CastEnd = endTime / 1000
 self.CastDuration = (endTime - startTime) / 1000

 self.F.CastBar:Show()
	self.F.Cast:Show()
	self.F.CastIconOverlay:Show()
 self.F.CastIconOverlay:SetTexture(icon)
 self.F.CastIconOverlay:SetTexCoord(0.1, 0.9, 0.1, 0.9)

 self.F.UnitName:Hide()
	self.F.NameSpannerL:Hide()
	self.F.NameSpannerS:Hide()
	self.F.NameSpannerR:Hide()

end

function SkyUnitFrame.prototype:CastUpdate()
 if not self.Spell then return end
	local p, v, t = (self.CastEnd - GetTime()) / self.CastDuration, self.T.Cast, self.F.Cast
 if p > 1 then p = 1 end

	if self.CastIsChanneled then
	
	else p = 1 - p end
	
 local newWidth = v.W * p
	local dif = v.W - newWidth

 local g = self.TS
	

	local x1, x2, y1, y2 = v.tX / g.W, (v.tX + newWidth) / g.W, v.tY / g.H, (v.tY + v.H) / g.H

 point, relativeTo, relativePoint, xOfs, yOfs = t:GetPoint()

	if self.TexturesReversed then
	 t:SetTexCoord(x2, y1, x2, y2, x1, y1, x1, y2)
 	t:SetPoint(relativePoint, t:GetParent(), -v.X - self.tO.X, v.Y + self.tO.Y)
	else
	 t:SetTexCoord(x1, x2, y1, y2)
 	t:SetPoint(relativePoint, t:GetParent(), v.X + self.tO.X, v.Y + self.tO.Y)
	end
	
	t:SetSize(newWidth, v.H)

	
end

function SkyUnitFrame.prototype:CastStop(evt, ...)

 self.IsCasting = false
 self.Spell = nil
	self.CastEnd = nil
 self.CastDuration = nil
	self.CastIsChanneled = false;

 self.F.CastBar:Hide()
	self.F.Cast:Hide()
	self.F.CastIconOverlay:Hide()

 self.F.UnitName:Show()
	self.F.NameSpannerL:Show()
	self.F.NameSpannerS:Show()
	self.F.NameSpannerR:Show()
end


function SkyUnitFrame.prototype:IncomingHealsUpdate()

	self.IncomingHeals = UnitGetIncomingHeals(self.Unit) or 0
 if self.IncomingHeals ~= 0 then self.IncomingHeals = self.IncomingHeals + UnitHealth(self.Unit) end
	if self.IncomingHeals >= UnitHealthMax(self.Unit) then self.IncomingHeals = 0 end
 self:UpdateVertBar(self.F.IncomingHealth, self.IncomingHeals, UnitHealthMax(self.Unit), self.T.IncomingHealth)
  
end


local function compareDurations(a,b)
 if not a[5] or not b[5] then return a[1] < b[1]
 else return a[5] > b[5] end
end


function SkyUnitFrame.prototype:UpdateAuraBars() -- ! ** !  ! ** !  ! ** !  ! ** !  ! ** !  ! ** !  ! ** !  ! ** !  ! ** !  ! ** !  ! ** !  ! ** !  ! ** !  ! ** !  --
 
-- for k, v in pairs(self.AuraBars) do
-- local str = ""
 for i = 1, self.AuraBarCount - 1 do
  v = self.AuraBars[i]
  local durationRemaining = v.Expires - GetTime()
--  if string.sub(v.Bar, 1, 1) == "D" then str = str .. v.Bar .. ", " .. v.Duration .. ", " .. durationRemaining .. "\n" end

  local dur = ""
  
  local dur = string.format("%.1f", durationRemaining)


  local c = cfg.Auras[v.cfg]
  
  if c.Names then
   if c.CD then
    if self.TexturesReversed then dur = v.Spell .. " " .. dur
    else dur = dur .. " " .. v.Spell end
   else
    dur = v.Spell
   end
  elseif not c.CD then dur = "" end
  

  self.F[v.Txt .. "txt"]:SetText(dur)
  self.F[v.Txt .. "txtBG"]:SetWidth(self.F[v.Txt .. "txt"]:GetStringWidth())

  if cfg.Auras.CooldownBackdrop then self.F[v.Txt .. "txtBG"]:Show()
  else self.F[v.Txt .. "txtBG"]:Hide() end

  if c.Bars then self:UpdateVertBar(self.F[v.Bar], durationRemaining, v.Duration, v.T, v.Offset) end
 end
end



function SkyUnitFrame.prototype:AuraUpdates()
 if not cfg.Auras.Enabled then return end
-- local buffs, debuffs, aurabuffs, auradebuffs = {}, {}, {}, {} -- 

 local auras = {}
 local bmax, dmax = cfg.Auras.Buffs.Max, cfg.Auras.Debuffs.Max
 local harm, help = "HARMFUL", "HELPFUL"

 if cfg.Auras.Swap then harm, help = help, harm end
 
 local a, _ = 1, nil
 for i = 1, 20 do
	 bName, _, bIcon, _, bDebuffType, bDuration, bExpirationTime, _, bIsStealable, _, bSpellId = UnitAura(self.Unit, i, help) 
	 dName, _, dIcon, _, dDebuffType, dDuration, dExpirationTime, _, dIsStealable, _, dSpellId = UnitAura(self.Unit, i, harm) 
  if bName then
   if cfg.Auras.Buffs.Enabled then
	  if bDuration ~= 0 and bDuration < 16 then
	   auras[a] = { 1, bName, bIcon, bDuration, bExpirationTime, bIsStealable, bSpellId, i, "HELPFUL" } a = a + 1
	  else
	   auras[a] = { 2, bName, bIcon, bDuration, bExpirationTime, bIsStealable, bSpellId, i, "HELPFUL" } a = a + 1
	  end
	 end
  end

  if dName then
   if cfg.Auras.Debuffs.Enabled then
    if dDuration ~= 0 and dDuration < 16 then
     auras[a] = { 3, dName, dIcon, dDuration, dExpirationTime, dIsStealable, dSpellId, i, "HARMFUL" } a = a + 1
    else
     auras[a] = { 4, dName, dIcon, dDuration, dExpirationTime, dIsStealable, dSpellId, i, "HARMFUL" } a = a + 1
    end
   end
  end

  if not bName and not dName then break end
 end
 
 table.sort(auras, compareDurations)


		
		local bb, b, db, d, k, j = 1, 1, 1, 1, nil, 1
		
--		for i, k in ipairs(auras) do
		for i = 1, a - 1 do
		 k = auras[i]
--		 print(i, k)
		 if k[1] == 1 and bb <= bmax and k[4] > 3 then
			 if self.F["BuffBar" .. bb] then
			 	
			 	if cfg.Auras.Buffs.Bars then
			 	 self.F["BuffBar" .. bb]:SetAlpha(1)
			 	 self.F["BuffLife" .. bb]:SetAlpha(1)
			 	else
         self.F["BuffBar" .. bb]:SetAlpha(0)
         self.F["BuffLife" .. bb]:SetAlpha(0)
			 	end
        self.F["BuffIconBG" .. bb]:SetAlpha(1)
			 	
        self.F["BuffFrame" .. bb]:Show()
        self.F["BuffFrame" .. bb].TT = { Index = k[8], Unit = self.Unit, Filter = k[9] }
			 	self.F["BuffFrame" .. bb].texture:SetTexture(k[3])
			 	
			 	self.AuraBars[j] = { Bar = "BuffLife" .. bb, Txt = "BuffFrame" .. bb, Spell = k[2], Duration = k[4], Expires = k[5], T = self.T.Buff, Offset = (bb - 1) * 17, cfg = "Buffs" }
			 	j = j + 1
			  bb = bb + 1
			 end
			elseif k[1] == 2 or bb > bmax or (k[1] < 3 and k[4] < 4) then
			 if self.F["AuraBuff" .. b] then
			  self.F["AuraBuff" .. b]:SetAlpha(1)
        self.F["AuraBuffFrame" .. b]:Show()
        self.F["AuraBuffFrame" .. b].TT = { Index = k[8], Unit = self.Unit, Filter = k[9] }
				self.F["AuraBuffFrame" .. b].texture:SetTexture(k[3])
			  b = b + 1
			 end
			elseif k[1] == 3 and db <= dmax and k[4] > 3 then
			 if self.F["DebuffBar" .. db] then

        if cfg.Auras.Debuffs.Bars then
			   self.F["DebuffBar" .. db]:SetAlpha(1)
			   self.F["DebuffLife" .. db]:SetAlpha(1)
			  else
         self.F["DebuffBar" .. db]:SetAlpha(0)
         self.F["DebuffLife" .. db]:SetAlpha(0)
        end
        self.F["DebuffIconBG" .. db]:SetAlpha(1)
        
        
        self.F["DebuffFrame" .. db]:Show()
        self.F["DebuffFrame" .. db].TT = { Index = k[8], Unit = self.Unit, Filter = k[9] }
				self.F["DebuffFrame" .. db].texture:SetTexture(k[3])
			  self.AuraBars[j] = { Bar = "DebuffLife" .. db, Txt = "DebuffFrame" .. db, Spell = k[2], Duration = k[4], Expires = k[5], T = self.T.Debuff, Offset = -(db - 1) * 17, cfg = "Debuffs" }
        j = j + 1
			  db = db + 1
			 end
			elseif k[1] == 4 or db > dmax or k[4] < 4 then
			 if self.F["AuraDebuff" .. d] then
			  self.F["AuraDebuff" .. d]:SetAlpha(1)
        self.F["AuraDebuffFrame" .. d]:Show()
        self.F["AuraDebuffFrame" .. d].TT = { Index = k[8], Unit = self.Unit, Filter = k[9] }
				self.F["AuraDebuffFrame" .. d].texture:SetTexture(k[3])
			  d = d + 1
			 end
			end
			
		end
		self.AuraBarCount = j

  for i = 1, 12 do
  
   if i >= bb then
    self.F["BuffBar" .. i]:SetAlpha(0)
    self.F["BuffLife" .. i]:SetAlpha(0)
    self.F["BuffFrame" .. i].Index = nil
    self.F["BuffFrame" .. i]:Hide()
    self.F["BuffIconBG" .. i]:SetAlpha(0)
   end
   if i >=db then
    self.F["DebuffBar" .. i]:SetAlpha(0)
    self.F["DebuffLife" .. i]:SetAlpha(0)
    self.F["DebuffFrame" .. i].Index = nil
    self.F["DebuffFrame" .. i]:Hide()
    self.F["DebuffIconBG" .. i]:SetAlpha(0)
   end
   if i >= b then
    self.F["AuraBuff" .. i]:SetAlpha(0)
    self.F["AuraBuffFrame" .. i].Index = nil
    self.F["AuraBuffFrame" .. i]:Hide()
   end
   if i >= d then
    self.F["AuraDebuff" .. i]:SetAlpha(0)
    self.F["AuraDebuffFrame" .. i].Index = nil
    self.F["AuraDebuffFrame" .. i]:Hide()
   end
  end
  


end

function SkyUnitFrame.prototype:UpdateClassColor()
 local class, classFileName = UnitClass(self.Unit)
 local c = RAID_CLASS_COLORS[classFileName]
 if c then
  self.F.ClassColorStrip:SetVertexColor(c.r, c.g, c.b)
 end
end







function SkyUnitFrame.prototype:UpdateUnitName()
 self.F.UnitName:SetText(UnitName(self.Unit))
	local w = self.F.UnitName:GetStringWidth() + 6
	self:UpdateNameSpanner(w)
	self:UpdateLevel()
end

function SkyUnitFrame.prototype:UpdateLevel()
 local l = UnitLevel(self.Unit)
 if l == -1 then l = "??" end
 self.F.UnitLevel:SetText(l)
 local w = self.F.UnitLevel:GetStringWidth() / 2 - 13
 if self.TexturesReversed then self.F.UnitLevel:SetPoint("TOPRIGHT", -self.Txt.UnitLevel.X - self.tO.X + w, self.Txt.UnitLevel.Y + self.tO.Y)
 else self.F.UnitLevel:SetPoint("TOPLEFT", self.Txt.UnitLevel.X + self.tO.X - w + 2, self.Txt.UnitLevel.Y + self.tO.Y) end  
end




function SkyUnitFrame.prototype:UpdateNameSpanner(w)
 self.F.NameSpannerS:SetWidth(w + 1)
 if self.TexturesReversed then self.F.NameSpannerR:SetPoint("TOPRIGHT", -self.T.NameSpannerR.X - self.tO.X - w, self.T.NameSpannerR.Y + self.tO.Y)
 else self.F.NameSpannerR:SetPoint("TOPLEFT", self.T.NameSpannerR.X + self.tO.X + w, self.T.NameSpannerR.Y + self.tO.Y) end
end



function SkyUnitFrame.prototype:PLAYER_ENTERING_WORLD()
		self:UpdateVertBar(self.Health, UnitHealth(self.Unit), UnitHealthMax(self.Unit), self.T.Health)
end

function SkyUnitFrame.prototype:UpdateDisplayPower()

  local powerType, powerToken, _, _, _ = select(1, UnitPowerType(self.Unit))
  if not powerType or (powerType > 0 and powerType < 15) or UnitPowerMax(self.Unit) == 100 then self.F.UnitPowerPercent:SetAlpha(0)
  else self.F.UnitPowerPercent:SetAlpha(1) end
  

  if UnitPowerMax(self.Unit) == 0 then
   self.F.Power:SetAlpha(0)
   self.F.UnitPower:SetVertexColor(1, 1, 1, 1)
  else
   self.F.Power:SetVertexColor(SkyUI.PBC[powerType].r, SkyUI.PBC[powerType].g, SkyUI.PBC[powerType].b, 1)
   self.F.UnitPower:SetVertexColor(SkyUI.PBC[powerType].rT, SkyUI.PBC[powerType].gT, SkyUI.PBC[powerType].bT, 1)
  end
--  if UnitExists(self.Unit .. "pet") then 
--   local powerType = select(1, UnitPowerType(self.Unit .. "pet"))
--   self.F.PetPower:SetVertexColor(PowerBarColor[powerType].r, PowerBarColor[powerType].g, PowerBarColor[powerType].b)
--  end  
  
end



function SkyUnitFrame.prototype:Pet_Changed()
 if cfg[self.Unit .. "pet"].Enabled and UnitName(self.Unit .. "pet") then
  self:SetText_Justified("PetName", UnitName(self.Unit .. "pet"), nil, "RIGHT", 82)
  self:SetText_Justified("PetLvl", UnitLevel(self.Unit .. "pet"), nil, "CENTER")
  
  if not self.F.PetBG:IsShown() then
   self.F.PetBG:Show()
   self.F.PetHealth:Show()
   self.F.PetPower:Show()
   self.F.PetName:Show()
   self.F.PetPortrait:Show()
   self.F.PetLvl:Show()
   self.F.PetLvlBG:Show()
   EnableClickable(self, self.Unit .. "pet")
  end
  
  SetPortraitTexture(self.F.PetPortrait, self.Unit .. "pet")
  
  local powerType = select(1, UnitPowerType(self.Unit .. "pet"))
  self.F.PetPower:SetVertexColor(SkyUI.PBC[powerType].r, SkyUI.PBC[powerType].g, SkyUI.PBC[powerType].b)
  
  if not powerType or UnitPowerMax(self.Unit .. "pet") == 0 then self.F.PetPower:SetAlpha(0)
  else self.F.PetPower:SetAlpha(1) end

--  print("Portrait Updated->" .. caller)
 elseif self.F.PetBG:IsShown() then
  self.F.PetBG:Hide()
  self.F.PetHealth:Hide()
  self.F.PetPower:Hide()
  self.F.PetName:Hide()
  self.F.PetPortrait:Hide()
  self.F.PetLvl:Hide()
  self.F.PetLvlBG:Hide()
  DisableClickable(self, self.Unit .. "pet")
 end
 self:PetTarget_Changed()
  
end

function SkyUnitFrame.prototype:PetPortrait_Update()
end

function SkyUnitFrame.prototype:PetTarget_Changed()
 if cfg[self.Unit .. "pettarget"].Enabled and UnitExists(self.Unit .. "pettarget") and UnitName(self.Unit .. "pettarget") then
  self:SetText_Justified("PetTargetName", UnitName(self.Unit .. "pettarget"), nil, "RIGHT", 82)
  self:SetText_Justified("PetTargetLvl", UnitLevel(self.Unit .. "pettarget"), nil, "CENTER")

  if not self.F.PetTargetBG:IsShown() then
   self.F.PetTargetBG:Show()
   self.F.PetTargetHealth:Show()
   self.F.PetTargetPower:Show()
   self.F.PetTargetName:Show()
   self.F.PetTargetPortrait:Show()
   self.F.PetTargetLvl:Show()
   self.F.PetTargetLvlBG:Show()
   EnableClickable(self, self.Unit .. "pettarget")
  end
  
  local powerType = select(1, UnitPowerType(self.Unit .. "pettarget"))
  self.F.PetTargetPower:SetVertexColor(PowerBarColor[powerType].r, PowerBarColor[powerType].g, PowerBarColor[powerType].b)
  
  if not powerType or UnitPowerMax(self.Unit .. "pettarget") == 0 then self.F.PetTargetPower:SetAlpha(0)
  else self.F.PetTargetPower:SetAlpha(1) end


  SetPortraitTexture(self.F.PetTargetPortrait, self.Unit .. "pettarget")
  
 elseif self.F.PetTargetBG:IsShown() then
  self.F.PetTargetBG:Hide()
  self.F.PetTargetHealth:Hide()
  self.F.PetTargetPower:Hide()
  self.F.PetTargetName:Hide()
  self.F.PetTargetPortrait:Hide()
  self.F.PetTargetLvl:Hide()
  self.F.PetTargetLvlBG:Hide()
  DisableClickable(self, self.Unit .. "pettarget")
 end
  
end 

function SkyUnitFrame.prototype:TargetOrFocus_Changed()
 local unit
 if self.Unit == "player" then unit = "focus"
 else unit = self.Unit .. "target" end
 
 if (cfg.targettarget.Enabled and self.Unit == "target" and UnitExists("targettarget")) or (cfg.focus.Enabled and self.Unit == "player" and UnitExists("focus")) then

  local name = select(1, UnitName(unit))
  if not name then name = "??" end

  local level = UnitLevel(unit)
  if level == -1 then level = "??" end

  if not self.F.TargetOrFocusBG:IsShown() then
   self.F.TargetOrFocusBG:Show()
   self.F.TargetOrFocusHealth:Show()  
   self.F.TargetOrFocusPower:Show()
   self.F.TargetOrFocusName:Show()
   self.F.TargetOrFocusPortrait:Show()
   self.F.TargetOrFocusLvl:Show()
   self.F.TargetOrFocusLvlBG:Show()
   EnableClickable(self, unit)
  end

  local powerType = select(1, UnitPowerType(unit))
  self.F.TargetOrFocusPower:SetVertexColor(PowerBarColor[powerType].r, PowerBarColor[powerType].g, PowerBarColor[powerType].b)
  
  if not powerType or UnitPowerMax(unit) == 0 then self.F.TargetOrFocusPower:SetAlpha(0)
  else self.F.TargetOrFocusPower:SetAlpha(1) end

  SetPortraitTexture(self.F.TargetOrFocusPortrait, unit)
  self:SetText_Justified("TargetOrFocusName", name, nil, "RIGHT", 82)
  self:SetText_Justified("TargetOrFocusLvl", level, nil, "CENTER")
  
 elseif self.F.TargetOrFocusBG:IsShown() then
  self.F.TargetOrFocusBG:Hide()
  self.F.TargetOrFocusHealth:Hide()
  self.F.TargetOrFocusPower:Hide()
  self.F.TargetOrFocusName:Hide()
  self.F.TargetOrFocusPortrait:Hide()
  self.F.TargetOrFocusLvl:Hide()
  self.F.TargetOrFocusLvlBG:Hide()
  DisableClickable(self, unit)
 end
 
 
end

function SkyUnitFrame.prototype:UpdateUnitFrames()
  self:Pet_Changed()
  self:TargetOrFocus_Changed()
end


-- Remove? --
function SkyUnitFrame.prototype:HideAuras()
 for i = 1, 12 do
  if i < 9 then
   self.F["BuffBar" .. i]:SetAlpha(0)
   self.F["BuffLife" .. i]:SetAlpha(0)
   self.F["BuffFrame" .. i].Index = nil
   self.F["BuffFrame" .. i]:Hide()
   self.F["DebuffBar" .. i]:SetAlpha(0)
   self.F["DebuffLife" .. i]:SetAlpha(0)
   self.F["DebuffFrame" .. i].Index = nil
   self.F["DebuffFrame" .. i]:Hide()
   self.F["BuffIconBG" .. i]:Hide()
   self.F["DebuffIconBG" .. i]:Hide()
  end
  self.F["AuraBuff" .. i]:SetAlpha(0)
  self.F["AuraBuffFrame" .. i].Index = nil
  self.F["AuraBuffFrame" .. i]:Hide()
  self.F["AuraDebuff" .. i]:SetAlpha(0)
  self.F["AuraDebuffFrame" .. i].Index = nil
  self.F["AuraDebuffFrame" .. i]:Hide()
 end
end

function SkyUnitFrame.prototype:Enable()
  self.F:SetScript("OnUpdate", self.Update)
  self:CastStart()
  self:AuraUpdates()
  if UnitExists(self.Unit) then SkyUnitFrame.super.prototype.Enable(self) end
end

function SkyUnitFrame.prototype:Disable()
  self:CastStop()
  self.F:SetScript("OnUpdate", nil)
  SkyUnitFrame.super.prototype.Disable(self)
end

function SkyUnitFrame.prototype:OnConfigUpdated()
 local frameSpec = SkyUI.Specs.UnitFrames.Frames[self.Unit:gsub("^%l", string.upper)]
 local newX, newY = cfg.UI.OffsetX, cfg.UI.OffsetY
 if self.Unit == "player" then
  self.F:SetPoint(self.RT, frameSpec.X + newX, frameSpec.Y + newY)
  self.F:SetAlpha(cfg.UI.player.OOCAlpha)
 else
  self.F:SetPoint(self.RT, frameSpec.X - newX, frameSpec.Y + newY)
  
 end
 self:UpdateUnitName()
 self:UpdateClassColor()
 self:UpdateDisplayPower()
 self:UpdateUnitFrames()
 self:AuraUpdates()
end


local function EventManager(self, evt, unit, arg1, arg2, arg3)
if unit == nil then unit = "nil" end
 if evt == "UNIT_PORTRAIT_UPDATE" then
  if unit == "pet" then SkyUI.player:Pet_Changed() SkyUI.target:Pet_Changed()
  elseif unit == "player" then SetPortraitTexture(SkyUI.player.F.Portrait, "player") SkyUI.player:UpdateUnitName()end
--  if unit == "player" then 
--  elseif unit == "target" then
 elseif evt == "UNIT_DISPLAYPOWER" then
  if unit == "player" then SkyUI.player:UpdateDisplayPower()
  elseif unit == "target" then SkyUI.target:UpdateDisplayPower()
  elseif unit == "pet" then
   SkyUI.player:Pet_Changed()
   SkyUI.target:Pet_Changed()
  end
 
 elseif evt == "PLAYER_FOCUS_CHANGED" then SkyUI.player:TargetOrFocus_Changed()
-- elseif evt == "PLAYER_TARGET_CHANGED" then SkyUI.target:PlayerTarget_Changed()
 elseif evt == "UNIT_TARGET" then
  if unit == "pet" then
--   for key,value in pairs(self) do print(key,value) end
   SkyUI.player:PetTarget_Changed()
   SkyUI.target:PetTarget_Changed()
  elseif unit == "target" then SkyUI.target:TargetOrFocus_Changed() SkyUI.target:Pet_Changed()
  end
--  if UnitGUID("target") then SkyUI.target:PLAYER_TARGET_CHANGED() end 

 elseif evt == "PLAYER_ENTERING_WORLD" or evt == "LEARNED_SPELL_IN_TAB" then
  UpdateSpellList()
 elseif evt == "PLAYER_XP_UPDATE" then
 SkyUI.player:UpdateUnitName()
	elseif SkyUI[unit] then
--  if evt == "UNIT_DISPLAYPOWER" then SkyUI[unit]:UNIT_DISPLAYPOWER()
  if evt == "UNIT_PET" then
   if unit == "target" then SkyUI.target:Pet_Changed()
   elseif unit == "player" then SkyUI.player:Pet_Changed() end  
  elseif evt == "UNIT_SPELLCAST_START" or evt == "UNIT_SPELLCAST_CHANNEL_START" then SkyUI[unit]:CastStart(evt, unit)
  elseif evt == "UNIT_SPELLCAST_STOP" or evt == "UNIT_SPELLCAST_CHANNEL_STOP" then SkyUI[unit]:CastStop(evt, unit)
	
	 elseif evt == "UNIT_HEAL_PREDICTION" then SkyUI[unit]:IncomingHealsUpdate(evt, unit)
 
  elseif evt == "UNIT_AURA" and SkyUI[unit] then SkyUI[unit]:AuraUpdates()
 
  elseif evt == "UNIT_THREAT_SITUATION_UPDATE" then
   SkyUI.target:UpdateUnitReaction()
 end

 elseif evt == "PLAYER_REGEN_ENABLED" then
  UIFrameFadeOut(SkyUI.player.F, 1.25, cfg.UI.player.CombatAlpha, cfg.UI.player.OOCAlpha)
 elseif evt == "PLAYER_REGEN_DISABLED" then
  SkyUI.player.F:SetAlpha(cfg.UI.player.CombatAlpha)
 end
end

local evt


SkyUI.Modules.UnitFrames.OnConfigUpdated = function()
 if SkyUI.Modules.UnitFrames.Enabled == false then return true end
 SkyUI.player:OnConfigUpdated()
 SkyUI.target:OnConfigUpdated()
 
 
end





SkyUI.Modules.UnitFrames.Auras.Disable = function()
 if not SkyUI.Modules.UnitFrames.Auras.Enabled then return true end
 cfg.Auras.Enabled = false
 SkyUI.Modules.UnitFrames.Auras.Enabled = false

-- SkyUI.player:HideAuras()
-- SkyUI.target:HideAuras()

end

SkyUI.Modules.UnitFrames.Auras.Enable = function()
 if SkyUI.Modules.UnitFrames.Auras.Enabled then return true end
 cfg.Auras.Enabled = true
 SkyUI.Modules.UnitFrames.Auras.Enabled = true

 SkyUI.player:AuraUpdates()
 SkyUI.target:AuraUpdates()
end




SkyUI.Modules.UnitFrames.Enable = function()
 if SkyUI.Modules.UnitFrames.Enabled == true then return true end
  
 SkyUI.Modules.UnitFrames.Enabled = true
 cfg.UI.Enabled = true
 
 if not evt then evt = CreateFrame("Frame") end

 if SkyUI.player then SkyUI.player:Enable()
 else SkyUnitFrame:new("player") end

 if SkyUI.target then SkyUI.target:Enable()
 else SkyUnitFrame:new("target") end

 RegisterStateDriver(SkyUI.player.F, "visibility", "[petbattle] hide; show")
 RegisterStateDriver(SkyUI.target.F, "visibility", "[petbattle] hide; show")


 UpdateSpellList()
-- debug = evt:CreateFontString("debugString", "OVERLAY")
-- debug:SetFont(SkyUI.Fonts.Marken, 10, "OUTLINE")
-- debug:ClearAllPoints()
-- debug:SetPoint("TOPLEFT", UIParent, "TOPLEFT", 25, -25)
-- debug:SetText("")


 


 RegisterStateDriver(PlayerFrame, "visibility", "hide")
 RegisterStateDriver(TargetFrame, "visibility", "hide")
 RegisterStateDriver(FocusFrame, "visibility", "hide")
 RegisterStateDriver(CastingBarFrame, "visibility", "hide")
  
  
--  TargetFrame:SetScript("OnEvent", nil);
--  TargetFrame:Hide();

-- evt:RegisterEvent("PLAYER_ENTERING_WORLD") !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
 evt:RegisterEvent("PLAYER_XP_UPDATE")

 evt:RegisterEvent("UNIT_DISPLAYPOWER")

 evt:RegisterEvent("PLAYER_TARGET_CHANGED")
 evt:RegisterEvent("UNIT_TARGET")
 evt:RegisterEvent("PLAYER_FOCUS_CHANGED")
 evt:RegisterEvent("UNIT_PET")

-- evt:RegisterEvent("UNIT_HEALTH")
-- evt:RegisterEvent("UNIT_POWER")


 evt:RegisterEvent("UNIT_SPELLCAST_START")
 evt:RegisterEvent("UNIT_SPELLCAST_STOP")

--evt:RegisterEvent("UNIT_SPELLCAST_FAILED")
--evt:RegisterEvent("UNIT_SPELLCAST_INTERRUPTED")

--evt:RegisterEvent("UNIT_SPELLCAST_DELAYED")
--evt:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")

 evt:RegisterEvent("UNIT_SPELLCAST_CHANNEL_START")
--evt:RegisterEvent("UNIT_SPELLCAST_CHANNEL_UPDATE")
 evt:RegisterEvent("UNIT_SPELLCAST_CHANNEL_STOP")

 evt:RegisterEvent("UNIT_HEAL_PREDICTION")

 evt:RegisterEvent("UNIT_AURA")

 evt:RegisterEvent("UNIT_THREAT_SITUATION_UPDATE")

 evt:RegisterEvent("UNIT_PORTRAIT_UPDATE")
 evt:RegisterEvent("LEARNED_SPELL_IN_TAB");

 evt:RegisterEvent("PLAYER_REGEN_ENABLED")
 evt:RegisterEvent("PLAYER_REGEN_DISABLED")

 evt:SetScript("OnEvent", EventManager)


end



SkyUI.Modules.UnitFrames.Disable = function()
 if SkyUI.Modules.UnitFrames.Enabled == false then return true end
 
 if SkyUI.player then SkyUI.player:Disable() end
 if SkyUI.target then SkyUI.target:Disable() end

-- UnregisterStateDriver(SkyUI.player.F, "visibility")
-- UnregisterStateDriver(SkyUI.target.F, "visibility")
 UnregisterStateDriver(SkyUI.player.F, "visibility")
 UnregisterStateDriver(SkyUI.target.F, "visibility")


 RegisterStateDriver(PlayerFrame, "visibility", "show")
 UnregisterStateDriver(PlayerFrame, "visibility")
 UnregisterStateDriver(TargetFrame, "visibility")
 UnregisterStateDriver(FocusFrame, "visibility")
 UnregisterStateDriver(CastingBarFrame, "visibility")
 PlayerFrame.Show = nil
 
-- PetFrame

 if UnitExists("target") then TargetFrame:Show() end
 if UnitExists("focus") then FocusFrame:Show() end

 evt:UnregisterEvent("PLAYER_XP_UPDATE")
 evt:UnregisterEvent("UNIT_DISPLAYPOWER")
 evt:UnregisterEvent("PLAYER_TARGET_CHANGED")
 evt:UnregisterEvent("UNIT_TARGET")
 evt:UnregisterEvent("PLAYER_FOCUS_CHANGED")
 evt:UnregisterEvent("UNIT_PET")
 evt:UnregisterEvent("UNIT_SPELLCAST_START")
 evt:UnregisterEvent("UNIT_SPELLCAST_STOP")
 evt:UnregisterEvent("UNIT_SPELLCAST_CHANNEL_START")
 evt:UnregisterEvent("UNIT_SPELLCAST_CHANNEL_STOP")
 evt:UnregisterEvent("UNIT_HEAL_PREDICTION")
 evt:UnregisterEvent("UNIT_AURA")
 evt:UnregisterEvent("UNIT_THREAT_SITUATION_UPDATE")
 evt:UnregisterEvent("UNIT_PORTRAIT_UPDATE")

 evt:UnregisterEvent("PLAYER_REGEN_ENABLED")
 evt:UnregisterEvent("PLAYER_REGEN_DISABLED")

 evt:SetScript("OnEvent", nil)

 SkyUI.Modules.UnitFrames.Enabled = false
 cfg.UI.Enabled = false
end





